// Copyright (c) 2015, the package authors. Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// MIT license that can be found in the LICENSE file.

library dartiny_event_dispatcher.test.event_dispatcher;

import 'package:test/test.dart';
import 'package:dartiny_event_dispatcher/dartiny_event_dispatcher.dart';

/// Names of events for testing.
class TestEvents {
  static const testEvent = 'test_event';
  static const anotherEvent = 'another_events';
}

/// Creates mock for listener.
EventListener getListenerMock(String listenerId, List<String> callsQueue,
                                 {String expectedEventName, Event expectedEvent, bool stopsPropagation: false}) {
  return
    (String eventName, Event event) async {
      if (expectedEventName != null) {
        expect(eventName, equals(expectedEventName));
      } else {
        expect(eventName, allOf(isNotNull, isNotEmpty));
      }

      if (expectedEvent != null) {
        expect(event, same(expectedEvent));
      } else {
        expect(event, const isInstanceOf<Event>());
      }

      if (stopsPropagation) {
        event.stopPropagation();
      }
      callsQueue.add(listenerId);
    };
}

void main() {

  group('EventDispatcher', () {

    EventDispatcherBuilder builder;

    setUp(() => builder = new EventDispatcherBuilder());
    tearDown(() => builder = null);

    group('.dispatch()', () {
      List<String> calls;

      setUp(() => calls = new List<String>());
      tearDown(() => calls = null);

      test('No listeners registered', () async {
        var dispatcher = builder.build();
        var event = new Event();

        var returned = await dispatcher.dispatch('some_event', event);
        expect(returned, same(event));
      });

      test('Event object is not passed', () async {
        builder.addListener(
            TestEvents.testEvent,
            getListenerMock('A', calls, expectedEventName: TestEvents.testEvent)
        );

        var dispatcher = builder.build();
        var returned = await dispatcher.dispatch(TestEvents.testEvent);

        expect(calls, ['A']);
        expect(returned, const isInstanceOf<Event>());
      });

      test('Event object is passed', () async {
        var event = new Event();

        builder.addListener(
            TestEvents.testEvent,
            getListenerMock('A', calls, expectedEventName: TestEvents.testEvent, expectedEvent: event)
        );

        var dispatcher = builder.build();
        var returned = await dispatcher.dispatch(TestEvents.testEvent, event);

        expect(calls, ['A']);
        expect(returned, same(event));
      });

      test('Listeners with identical priority', () async {
        builder
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('A', calls, expectedEventName: TestEvents.testEvent)
          )
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('B', calls, expectedEventName: TestEvents.testEvent)
          )
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('C', calls, expectedEventName: TestEvents.testEvent)
          )
          // Should not be called:
          ..addListener(
            TestEvents.anotherEvent,
            getListenerMock('D', calls)
          )
        ;

        var dispatcher = builder.build();
        await dispatcher.dispatch(TestEvents.testEvent);

        expect(calls, ['A', 'B', 'C']);
      });

      test('Listeners with different priority', () async {
        builder
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('A', calls, expectedEventName: TestEvents.testEvent),
            priority: 0
          )
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('B', calls, expectedEventName: TestEvents.testEvent),
            priority: 100
          )
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('C', calls, expectedEventName: TestEvents.testEvent),
            priority: 200
        )
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('D', calls, expectedEventName: TestEvents.testEvent),
            priority: 100
          )

          ..addListener(
            TestEvents.testEvent,
            getListenerMock('E', calls, expectedEventName: TestEvents.testEvent),
            priority: 200
          )
        ;

        var dispatcher = builder.build();
        await dispatcher.dispatch(TestEvents.testEvent);

        expect(calls, ['C', 'E', 'B', 'D', 'A']);
      });

      test('With initialy stopped propagation', () async {
        var event = new Event()..stopPropagation();

        builder.addListener(
          TestEvents.anotherEvent,
          getListenerMock('Never called', calls)
        );

        var dispatcher = builder.build();
        await dispatcher.dispatch(TestEvents.testEvent, event);

        expect(calls, []);
      });

      test('Some listener stops propagation', () async {
        builder
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('A', calls, expectedEventName: TestEvents.testEvent)
          )
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('B', calls, expectedEventName: TestEvents.testEvent, stopsPropagation: true)
          )
          ..addListener(
            TestEvents.testEvent,
            getListenerMock('C', calls, expectedEventName: TestEvents.testEvent)
          )
        ;
        var dispatcher = builder.build();
        var returned = await dispatcher.dispatch(TestEvents.testEvent);

        expect(returned.isPropagationStopped, isTrue);
        expect(calls, ['A', 'B']);
      });

    });

  });

}